class ExecutorError(Exception):
    pass


class StopPlugin(ExecutorError):
    def __init__(self) -> None:
        self.msg = "Plugin execution is stopped by user."
        super().__init__(self.msg)


class ExecutionError(ExecutorError):
    def __init__(self, plugin_name: str) -> None:
        self.plugin_name = plugin_name
        self.msg = f"During of the execution Plugin: <{plugin_name}> raised the error."
        super().__init__(self.msg)



from __future__ import annotations

import asyncio
import shelve
from functools import partial
from typing import TypeVar
from typing import TypedDict
from pydantic import SecretStr
from .types import (
    TesterID,
    EProductType
)

__all__ = ("PrecisionStorage",)


class Methods:
    @staticmethod
    def save(open_db: partial[shelve.Shelf], params: StorageResource) -> None:
        with open_db() as db:
            db[params["id"]] = params

    @staticmethod
    def get_all(open_db: partial[shelve.Shelf]) -> tuple[StorageResource]:
        with open_db() as db:
            return tuple(db.values())

    @staticmethod
    def delete(open_db: partial[shelve.Shelf], id: TesterID) -> None:
        with open_db() as db:
            if id not in db:
                return None
            del db[id]

    @staticmethod
    def is_registered(open_db: partial[shelve.Shelf], id: TesterID) -> bool:
        with open_db() as db:
            return id in db


T = TypeVar("T")


class StorageResource(TypedDict):
    id: TesterID
    product: EProductType
    host: str
    port: int
    password: SecretStr
    name: str
    keep_disconnected: bool


class PrecisionStorage:
    __slots__ = ("_lock", "_loop", "__open",)

    def __init__(self, storage_path: str) -> None:
        self._lock = asyncio.Lock()
        self._loop = asyncio.get_event_loop()
        self.__open = partial(shelve.open, storage_path)

    async def __run(self, func: partial[T]) -> T:
        async with self._lock:
            return await self._loop.run_in_executor(None, func)

    async def get_all(self) -> tuple[StorageResource]:
        method = partial(Methods.get_all, self.__open)
        return await self.__run(method)

    async def is_registered(self, t_id: TesterID) -> bool:
        method = partial(Methods.is_registered, self.__open, t_id)
        return await self.__run(method)

    async def save(self, params: StorageResource) -> None:
        method = partial(Methods.save, self.__open, params)
        return await self.__run(method)

    async def delete(self, t_id: TesterID) -> None:
        method = partial(Methods.delete, self.__open, t_id)
        return await self.__run(method)



import asyncio
from . import exceptions


class StateConditionsFacade:
    __slots__ = ("__wait_if_paused", "__stop_if_stopped")

    def __init__(self, wait_if_paused, stop_if_stopped) -> None:  # TODO: Add type notations
        self.__wait_if_paused = wait_if_paused
        self.__stop_if_stopped = stop_if_stopped

    async def wait_if_paused(self) -> None:
        await self.__wait_if_paused()

    async def stop_if_stopped(self) -> None:
        await self.__stop_if_stopped()


class StateConditions:
    __slots__ = ("continue_event", "stop_event")

    def __init__(self) -> None:
        self.continue_event = asyncio.Event()
        self.continue_event.set()
        self.stop_event = asyncio.Event()

    async def wait_if_paused(self) -> None:
        """Wait till user toggle pause state."""
        await self.continue_event.wait()

    async def stop_if_stopped(self) -> None:
        if self.stop_event.is_set():
            raise exceptions.StopPlugin()

    def get_facade(self) -> StateConditionsFacade:
        return StateConditionsFacade(
            self.wait_if_paused,
            self.stop_if_stopped
        )

    def pause(self) -> None:
        self.continue_event.clear()

    def resume(self) -> None:
        self.continue_event.set()

    def stop(self) -> None:
        self.stop_event.set()



from __future__ import annotations

import asyncio
from typing import (
    Any,
    Callable,
    Generator,
)

from pydantic import BaseModel

from .resource import const
from .resource.facade import Resource
from .resource.models.types import TesterID
from .resource.models.tester import TesterInfoModel
from .resource.exceptions import UnknownResourceError


""""
Messages Order
CONNECTED - Ignored message

1. ADDED - add tester to the pool send message to the user
    - CONNECTED - If know tester get connected send msg to user
    - CHANGED - If cnown tester data was changet sending msg to user
    - DISCONNECTED - If know tester get disconnected send msg to user
2. REMOVED - Extract tester and send message to the user

DISCONNECTED - Ignored message
"""


class MultiResActions:
    """
    Interface of actions applied to
    multiple Resources at the same time
    """

    __slots__ = ("resources",)

    def __init__(self, resources: dict[TesterID, Resource]) -> None:
        self.resources = resources

    async def connect(self) -> Generator[Resource, None, None]:
        prefiltered = (r for r in self.resources.values() if not r.keep_disconnected)
        result = await asyncio.gather(
            *[r.connect() for r in prefiltered],
            return_exceptions=True
        )
        return (
            resource
            for result, resource in zip(result, prefiltered)
            if isinstance(result, Exception)
        )

    def get_items(self) -> Generator[TesterInfoModel, None, None]:
        return (r.info() for r in self.resources.values())

    def select(self, tester_ids: tuple[TesterID, ...]) -> Generator[Resource, None, None]:
        for tester_id in tester_ids:
            res = self.resources.get(tester_id, None)
            if not res:
                raise UnknownResourceError(tester_id)
            yield res


class Msg(BaseModel):
    action: str
    data: TesterInfoModel


class ResourcesPool:
    __slots__ = ("__resources", "__publisher")

    def __init__(self, publisher: Callable[[Any], None]) -> None:
        self.__publisher = publisher
        self.__resources: dict[TesterID, Resource] = dict()

    def __contains__(self, key: TesterID) -> bool:
        return key in self.__resources

    def __len__(self) -> int:
        return len(self.__resources)

    def __optimize(self) -> None:
        """Realocate pool for keep low memory usage."""
        self.__resources = {**self.__resources}

    async def __publish_message(self, dataset: TesterInfoModel, event: str) -> None:
        message = Msg(action=event, data=dataset)
        self.__publisher(message)

    async def add(self, resource: Resource) -> None:
        """Add Resource to the pool and subscribe on changes"""
        self.__resources[resource.id] = resource
        self.__optimize()
        await self.__publish_message(resource.info(), const.ADDED)
        resource.events.on_changed(self.__publish_message)
        resource.events.on_connected(self.__publish_message)
        resource.events.on_disconnected(self.__publish_message)

    def get(self, id: TesterID) -> Resource:
        """Get a known Resource by it's ID"""
        if res := self.__resources.get(id, None):
            return res
        raise UnknownResourceError(id)

    async def extract(self, id: TesterID) -> Resource:
        """Exclude the Resource from the pool and stop propagate changes"""
        if resource := self.__resources.pop(id, None):
            self.__optimize()
            resource.events.reset()
            await self.__publish_message(resource.info(), const.REMOVED)
            return resource
        raise UnknownResourceError(id)

    @property
    def all(self) -> MultiResActions:
        return MultiResActions(self.__resources)



import asyncio
import contextlib
import uuid
import typing
from xoa_core.core.generic_types import (
    TObserver,
    TMesagesPipe,
)
if typing.TYPE_CHECKING:
    from xoa_core.types import PluginAbstract
    from xoa_core.core.plugin_abstract import (
        PStateConditionsFacade,
        PPipeFacade
    )

from .executor_info import ExecutorInfo
from . import exceptions
from ._events import Event
from .executor_state import ExecutorState
from .executor_state_conditions import StateConditions


class PPlugin(typing.Protocol):
    def create_test_suite(self, state_conditions: "PStateConditionsFacade", xoa_out: "PPipeFacade") -> "PluginAbstract": ...  # noqa: E704


class SuiteExecutor:
    __slots__ = ("suite_name", "state", "__id", "__observer", "__msg_pipe", "__test_suite", "__task", "state_conditions",)

    def __init__(self, suite_name: str) -> None:
        self.__id = str(uuid.uuid4())
        self.suite_name = suite_name
        self.state = ExecutorState()
        self.state_conditions = StateConditions()

    @property
    def id(self) -> str:
        return self.__id

    def __on_execution_terminated(self, task: "asyncio.Task") -> None:
        if not self.state.is_stoped:
            self.state.set_stop()
        err = None
        with contextlib.suppress(asyncio.CancelledError, exceptions.StopPlugin):
            err = task.exception()
            if err is not None:
                self.__msg_pipe.transmit_err(err)
                self.__observer.emit(Event.ERROR, task.get_name(), err)  # only notify of the Execution manager.
        self.__observer.emit(Event.STOPPED, self.id)
        asyncio.create_task(self.__msg_pipe.disable())  # In rare case raise error but it's not break anything.
        if err:
            raise exceptions.ExecutionError(task.get_name()) from err

    def get_info(self) -> ExecutorInfo:
        return ExecutorInfo(
            id=self.__id,
            suite_name=self.suite_name,
            state=self.state.current_state,
        )

    def assign_pipe(self, pipe: "TMesagesPipe") -> None:
        self.__msg_pipe = pipe
        self.state.assign_senders(pipe.get_state_facade())

    def assign_plugin(self, plugin: PPlugin) -> None:
        self.__test_suite = plugin.create_test_suite(
            state_conditions=self.state_conditions.get_facade(),
            xoa_out=self.__msg_pipe.get_facade(self.suite_name)
        )

    def run(self, observer: TObserver) -> None:
        self.__observer = observer
        self.state.set_run()
        self.__task = asyncio.create_task(
            self.__test_suite.start(),
            name=f"{self.suite_name}[{self.id}]"
        )
        self.__task.add_done_callback(self.__on_execution_terminated)

    async def toggle_pause(self) -> None:
        """User interface toggle pause."""
        if self.state.is_running:
            self.state.set_pause()
            self.state_conditions.pause()
            await self.__test_suite.on_pause()
        elif self.state.is_paused:
            self.state.set_run()
            self.state_conditions.resume()
            await self.__test_suite.on_continue()

    async def stop(self) -> None:
        """User interface stop the test suite."""
        self.state.set_stop()
        self.state_conditions.stop()
        await self.__test_suite.on_stop()
        self.__task.cancel()
        with contextlib.suppress(asyncio.CancelledError, exceptions.StopPlugin):
            await self.__task



from enum import IntEnum
from typing import (
    NewType,
    TypedDict,
)
from pydantic import SecretStr


TesterID = NewType("TesterID", str)
ModuleID = NewType("TesterID", str)  # <TesterID>-<Module Slot Index>
PortID = NewType("TesterID", str)  # <TesterID>-<ModuleID>-<Port Index>
JsonStr = NewType("JsonStr", str)


class EProductType(IntEnum):
    VALKYRIE = 1
    VULKAN = 2
    SAFIRE = 3
    VANTAGE = 4
    CHIMERA = 5


class StorageResource(TypedDict):
    id: TesterID
    product: EProductType
    host: str
    port: int
    password: SecretStr
    name: str
    keep_disconnected: bool



import functools
import typing


def post_notify(notifier: typing.Callable):
    def decorate(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            result = await func(*args, **kwargs)
            notifier()
            return result
        return wrapper
    return decorate



from xoa_core import __version__


class TestSuiteNotExistError(Exception):
    def __init__(self, name: str) -> None:
        self.name = name
        self.msg = f"Test Suite: <{self.name}> is not exist."
        super().__init__(self.msg)


class TestSuiteVersionError(Exception):
    def __init__(self, name: str, required_version: str) -> None:
        self.name = name
        self.required_version = required_version
        self.msg = f"Test Suite: <{self.name}> require framework version of <{self.required_version}> current is: <{__version__}>"
        super().__init__(self.msg)


class MultiModeError(Exception):
    def __init__(self) -> None:
        self.msg = "Only single Test suite execution is permited."
        super().__init__(self.msg)


class InvalidPluginError(ValueError):
    def __init__(self, value, expected) -> None:
        self.value = value
        self.expected = expected
        self.msg = f"Invalid plugin. Plugin Entry Class {self.value} must be a subclass of {self.expected}."
        super().__init__(self.msg)



from __future__ import annotations
import os
from pathlib import Path
import sys
import oyaml as yaml
import importlib.util
from typing import (
    TYPE_CHECKING,
    Optional,
    Type,
    Generator,
)

from pydantic import BaseModel
from xoa_core.core.exceptions import InvalidPluginError
if TYPE_CHECKING:
    from types import ModuleType

from .datasets import (
    PluginMeta,
    PluginData,
    build_test_params
)

from ..plugin_abstract import PluginAbstract


META_FILE_NAME = 'meta.yml'


def __load_module(path: str) -> Generator["ModuleType", None, None]:
    """Load module from path to the var"""
    for child in os.listdir(path):
        ilename, _ = os.path.splitext(child)
        module_name = f"{os.path.split(path)[-1]}.{ilename}"
        spec = importlib.util.spec_from_file_location(
            module_name,
            os.path.join(path, child),
        )
        if not spec or not spec.loader:
            continue
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        yield mod


def __make_plugin(module_path: str, meta: PluginMeta) -> Optional["PluginData"]:
    """Create plugin model."""
    entry_class: Optional[Type["PluginAbstract"]] = None
    model_class: Optional[Type["BaseModel"]] = None
    for module in __load_module(module_path):
        if entry_class is None:
            entry_class = getattr(module, meta.entry_object, None)
        if model_class is None:
            model_class = getattr(module, meta.data_model, None)
    if not (entry_class and model_class):
        return None
    if not issubclass(entry_class, PluginAbstract):
        raise InvalidPluginError(entry_class, PluginAbstract)
    if not issubclass(model_class, BaseModel):
        raise InvalidPluginError(model_class, BaseModel)
    return PluginData(
        meta=meta,
        entry_class=entry_class,
        model_class=build_test_params(model_class)
    )


def __read_meta(path: str) -> "PluginMeta":
    """Read metafile"""
    with open(path, "r") as outfile:
        data = yaml.load(outfile, Loader=yaml.SafeLoader)
        return PluginMeta(**data)

def __register_path(path: str | Path) -> None:
    str_path = str(path)
    if str_path in sys.path:
        return None
    sys.path.append(str_path)


def load_plugin(path: str | Path) -> Generator[PluginData, None, None]:
    __register_path(path)
    for child in os.listdir(path):
        child_path = os.path.abspath(os.path.join(path, child))
        if not os.path.isdir(child_path):
            continue
        meta_path = os.path.join(child_path, META_FILE_NAME)
        if not os.path.exists(meta_path):
            continue
        meta = __read_meta(meta_path)
        plugin = __make_plugin(child_path, meta)
        if not plugin:
            continue
        yield plugin



from __future__ import annotations
import asyncio
import contextlib
from functools import partialmethod
from typing import (
    Any,
    Final,
)

from xoa_core.core.generic_types import TObserver
from . import misc


class MesagesPipe:
    __slots__ = ("name", "__evt", "__queue", "__observer", "__push_streams", "__lock", "__procesor")

    def __init__(self, name: str, observer: "TObserver") -> None:
        self.name: Final[str] = name
        self.__evt = asyncio.Event()
        self.__queue: "asyncio.Queue[misc.Message]" = asyncio.Queue()
        self.__observer = observer
        self.__lock = asyncio.Lock()
        self.__push_streams: dict[str, asyncio.Queue["misc.Message" | None]] = {}
        self.__procesor = asyncio.create_task(
            self.__worker(),
            name=f"MessagesPipe[{self.name}]"
        )

    async def _add_stream(self, key: str, queue: asyncio.Queue) -> None:
        async with self.__lock:
            self.__push_streams[key] = queue

    async def _free_stream(self, key: str) -> None:
        async with self.__lock:
            self.__push_streams.pop(key)
            self.__push_streams = dict(self.__push_streams)

    async def __worker(self) -> None:
        while True:
            val = await self.__queue.get()
            async with self.__lock:
                await asyncio.gather(
                    *[
                        stm.put(val)
                        for stm in self.__push_streams.values()
                    ]
                )
            self.__queue.task_done()

    async def disable(self) -> None:
        self.__evt.set()
        await self.__queue.join()
        self.__procesor.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await self.__procesor
        for stm in self.__push_streams.values():
            stm.put_nowait(None)  # Inform to stop watching
        self.__observer.emit(misc.DISABLED, self.name)

    def transmit(self, msg: Any, *, msg_type: misc.EMsgType = misc.EMsgType.DATA, **meta: Any) -> None:
        """Unblocable function"""
        assert not self.__evt.is_set(), "Message pipe is closed"
        message = misc.Message(
            pipe_name=self.name,
            meta=meta,
            type=msg_type,
            payload=msg
        )
        self.__queue.put_nowait(message)

    def get_facade(self, suite_name: str) -> misc.PipeFacade:
        return misc.PipeFacade(self.transmit, suite_name)

    def get_state_facade(self) -> misc.PipeStateFacade:
        return misc.PipeStateFacade(self.transmit)

    transmit_warn = partialmethod(transmit, msg_type=misc.EMsgType.WARNING)
    transmit_err = partialmethod(transmit, msg_type=misc.EMsgType.ERROR)



from typing import (
    TYPE_CHECKING,
    Protocol,
    Any,
    Callable,
    # Optional,
    Coroutine,
    Tuple,
    AsyncGenerator,
    # TypeVar,
)

if TYPE_CHECKING:
    import asyncio
    # from xoa_driver import testers
    from functools import partialmethod
    # from valhalla_core.core.test_suites.datasets import PluginData

from xoa_core.core.messenger.misc import EMsgType, Message, PipeFacade, PipeStateFacade


# region Execution Manager

# TTesterReceaver = Callable[[str, str], Optional["testers.GenericAnyTester"]]

# endregion

# region Utils SimpleObserver

TCallback = Callable[..., Coroutine[Any, None, None]]


class TObserver(Protocol):
    def subscribe(self, evt: int, func: TCallback) -> None: ...  # noqa: E704
    def emit(self, evt: int, *args, **kwargs) -> None: ...  # noqa: E704

# endregion


# region Messanger


class TMesagesPipe(Protocol):
    async def _add_stream(self, key: str, queue: "asyncio.Queue") -> None: ...  # noqa: E704
    async def _free_stream(self, key: str) -> None: ...  # noqa: E704
    async def disable(self) -> None: ...  # noqa: E704
    def transmit(self, msg: Any, *, msg_type: EMsgType = EMsgType.DATA) -> None: ...  # noqa: E704
    def get_facade(self, suite_name: str) -> PipeFacade: ...  # noqa: E704
    def get_state_facade(self) -> PipeStateFacade: ...  # noqa: E704
    transmit_warn: "partialmethod"
    transmit_err: "partialmethod"


class TMessagesHandler(Protocol):
    def get_pipe(self, name: str) -> "TMesagesPipe": ...  # noqa: E704
    async def disable_pipe(self, name: str) -> None: ...  # noqa: E704
    def avaliable_pipes(self) -> Tuple[str, ...]: ...  # noqa: E704
    async def changes(self, *names: str) -> AsyncGenerator["Message", None]: ...  # noqa: E704


# endregion



import asyncio
import collections
from contextlib import suppress
from typing import (
    Dict,
    Generic,
    List,
    Callable,
    Any,
    Coroutine,
    TypeVar,
)
import warnings
CB = Callable[..., Coroutine[Any, None, None]]
T = TypeVar("T")


class SimpleObserver(Generic[T]):
    __slots__ = ("__events", "__pass_event")

    def __init__(self, *, pass_event: bool = False) -> None:
        self.__pass_event = pass_event
        self.reset()

    def __handle_exceptions(self, task: "asyncio.Task") -> None:
        with suppress(asyncio.CancelledError):
            if e := task.exception():
                warnings.warn(str(e))

    def reset(self) -> None:
        self.__events: Dict[T, List[CB]] = collections.defaultdict(list)

    def subscribe(self, evt: T, func: CB) -> None:
        self.__events[evt].append(func)

    def emit(self, evt: T, *args, **kwargs) -> None:
        kwargs_ = {**kwargs, "event": evt} if self.__pass_event else kwargs
        for action in self.__events[evt]:
            task = asyncio.create_task(action(*args, **kwargs_))
            task.add_done_callback(self.__handle_exceptions)



from __future__ import annotations

import typing
from pathlib import Path

from typing_extensions import Self

from .core import const
from .core.executors.executor import SuiteExecutor
from .core.executors.manager import ExecutorsManager
from .core.executors.executor_info import ExecutorInfo
from .core.messenger.handler import OutMessagesHandler
from .core.messenger.misc import Message
from .core.resources.controller import ResourcesController
from .core.resources.storage import PrecisionStorage
from .core.resources.types import Credentials, TesterInfoModel
from .core.test_suites.controller import PluginController
from .types import TesterID

if typing.TYPE_CHECKING:
    from .types import EMsgType


class MainController:
    """MainController - A main class of XOA-Core framework."""

    __slots__ = ("__is_started", "__publisher", "__resources", "suites_library", "__execution_manager")

    def __init__(self, *, storage_path: Path | str | None = None, mono: bool = False) -> None:
        self.__is_started = False
        __storage_path = Path.cwd() / "store" if not storage_path else Path(storage_path)

        self.__publisher = OutMessagesHandler()
        resources_pipe = self.__publisher.get_pipe(const.PIPE_RESOURCES)
        storage = PrecisionStorage(str(__storage_path))
        self.__resources = ResourcesController(resources_pipe, storage)

        executor_pipe = self.__publisher.get_pipe(const.PIPE_EXECUTOR)
        self.__execution_manager = ExecutorsManager(executor_pipe, mono)

        self.suites_library = PluginController()

    def listen_changes(self, *names: str, _filter: set["EMsgType"] | None = None) -> typing.AsyncGenerator[Message, None]:
        """Subscribe to the messages from different subsystems and test-suites."""
        return self.__publisher.changes(*names, _filter=_filter)

    def __await__(self) -> typing.Generator[typing.Any, None, Self]:
        return self.__setup().__await__()

    async def __setup(self) -> Self:
        if not self.__is_started:
            await self.__resources.start()
            self.__is_started = True
        return self

    def register_lib(self, path: str | Path) -> None:
        """Register lookup path of custom test suites library.

        :param path: lookup path of custom test suites library
        :type path: str | Path
        """
        self.suites_library.register_path(path)

    def get_available_test_suites(self) -> list[str]:
        """Get a list of available test suite names.

        :return: a list of available test suites
        :rtype: list[str]
        """
        return self.suites_library.available_test_suites()

    def get_test_suite_info(self, name: str) -> dict[str, typing.Any] | None:
        """Get the info of a test suite.

        :param name: name of the test suite
        :type name: str

        :return: dict of testsuite info or None
        :rtype: dict[str, typing.Any] | None
        """
        return self.suites_library.suite_info(name)

    async def list_testers_info(self) -> list[TesterInfoModel]:
        """List of info of known testers.

        :return: list of testers
        :rtype: list["TesterInfoModel"]
        """
        return await self.__resources.list_testers_info()

    async def get_tester_info(self, tester_id: TesterID) -> TesterInfoModel | None:
        """Info of an tester.

        :return: tester info
        :rtype: "TesterInfoModel"
        """
        return await self.__resources.get_tester_info(tester_id)

    async def add_tester(self, credentials: Credentials) -> TesterID:
        """Add a tester.

        :param credentials: tester login credentials
        :type credentials: credentials.Credentials
        :return: success or failure
        :rtype: bool
        """
        return await self.__resources.add_tester(credentials)

    async def remove_tester(self, tester_id: TesterID) -> None:
        """Remove a tester.

        :param tester_id: tester id
        :type tester_id: TesterID
        """
        await self.__resources.remove_tester(tester_id)

    async def connect_tester(self, tester_id: TesterID) -> None:
        """Establis connection to a disconnected tester.

        :param tester_id: tester id
        :type tester_id: TesterID
        """
        await self.__resources.connect(tester_id)

    async def disconnect_tester(self, tester_id: TesterID) -> None:
        """Disconnect from a tester.

        :param tester_id: tester id
        :type tester_id: TesterID
        """
        await self.__resources.disconnect(tester_id)

    def start_test_suite(self, test_suite_name: str, config: dict[str, typing.Any], *, debug_connection: bool = False) -> str:
        """Start test suite execution

        :param test_suite_name: test suite name
        :type test_suite_name: str
        :param config: test configuration data
        :type config: dict[str, typing.Any]
        :return: test execution id
        :rtype: str
        """
        plugin = self.suites_library.get_plugin(test_suite_name, debug_connection)
        plugin.parse_config(config)
        plugin.assign_testers(self.__resources.get_testers_by_id)
        executor = SuiteExecutor(test_suite_name)
        executor.assign_pipe(
            self.__publisher.get_pipe(executor.id)
        )
        executor.assign_plugin(plugin)
        return self.__execution_manager.run(executor)

    def executions_info(self) -> list[ExecutorInfo]:
        """
        Get list of infos of currently executing tasks
        :return: Executors info
        :rtype: list[ExecutorInfo]
        """
        return self.__execution_manager.get_executors_info()

    def execution_state(self, execution_id: str) -> str | None:
        """
        Get current state of the test execution

        :param execution_id: test execution id
        :type execution_id: str
        :return: str | none
        :rtype: str | None
        """
        return self.__execution_manager.get_state(execution_id)

    async def running_test_stop(self, execution_id: str) -> None:
        """Stop a test suite execution

        :param execution_id: test execution id
        :type execution_id: str
        :return: none
        :rtype: None
        """
        await self.__execution_manager.stop(execution_id)

    async def running_test_toggle_pause(self, execution_id: str) -> None:
        """Pause or continue execution of a test suite.

        :param execution_id: test execution id
        :type execution_id: str
        :return: none
        :rtype: None
        """
        return await self.__execution_manager.toggle_pause(execution_id)



from xoa_driver import testers
from xoa_driver import modules
from xoa_driver import ports


def is_tester(inst) -> bool:
    cmp_types = (
        testers.L23Tester,
        testers.L47Tester,
        testers.L47VeTester
    )
    return isinstance(inst, cmp_types)


def is_module(inst) -> bool:
    cmp_types = (
        modules.ModuleL23,
        modules.ModuleL23VE,
        modules.ModuleChimera,
        modules.ModuleL47,
        modules.ModuleL47VE,
    )
    return isinstance(inst, cmp_types)


def is_port(inst) -> bool:
    cmp_types = (
        ports.BasePortL23,
        ports.PortL23VE,
        ports.PortChimera,
        ports.PortL47,
    )
    return isinstance(inst, cmp_types)



from __future__ import annotations
from typing import (
    Dict,
    Optional,
    Protocol,
    Any,
)
from enum import Enum
from pydantic import BaseModel

DISABLED = 1


class EMsgType(Enum):
    STATE = "STATE"
    DATA = "DATA"
    STATISTICS = "STATISTICS"
    PROGRESS = "PROGRESS"
    WARNING = "WARNING"
    ERROR = "ERROR"


class Message(BaseModel):
    pipe_name: str
    meta: Dict[str, Any] = {}
    type: EMsgType = EMsgType.DATA
    payload: Any


class StatePayload(BaseModel):
    state: Optional[str]
    old_state: Optional[str]


class Progress(BaseModel):
    current: int
    total: int
    loop: int


class TransmitFunc(Protocol):
    def __call__(self, msg: Any, *, msg_type: EMsgType, **meta: Any) -> None:
        ...


class PipeStateFacade:
    __slots__ = ("__transmit",)

    def __init__(self, transmit: "TransmitFunc") -> None:
        self.__transmit = transmit

    def __call__(self, state: str | None, old_state: str | None) -> None:
        data = StatePayload(
            state=state,
            old_state=old_state
        )
        self.__transmit(data, msg_type=EMsgType.STATE)


class PipeFacade:
    __slots__ = ("__transmit", "__suite_name")

    def __init__(self, transmit: "TransmitFunc", suite_name: str) -> None:
        self.__transmit = transmit
        self.__suite_name = suite_name

    def send_statistics(self, data: dict[str, Any] | "BaseModel") -> None:
        self.__transmit(data, msg_type=EMsgType.STATISTICS, suite_name=self.__suite_name)

    def send_progress(self, current: int, total: int = 100, loop: int = 0) -> None:
        progress = Progress(
            current=current,
            total=total,
            loop=loop,
        )
        self.__transmit(progress, msg_type=EMsgType.PROGRESS, suite_name=self.__suite_name)

    def send_warning(self, warning: Exception) -> None:
        self.__transmit(str(warning), msg_type=EMsgType.WARNING, suite_name=self.__suite_name)

    def send_error(self, error: Exception) -> None:
        self.__transmit(str(error), msg_type=EMsgType.ERROR, suite_name=self.__suite_name)



from .core.plugin_abstract import PluginAbstract
from .core.test_suites.datasets import (
    PortIdentity,
    TestParameters
)

from .core.resources.types import (
    TesterID,
    EProductType,
    Credentials,
    TesterInfoModel,
    ModuleInfoModel,
    PortInfoModel,
)
from .core.messenger.misc import (
    EMsgType,
    Message,
)
from .core.const import (
    PIPE_EXECUTOR,
    PIPE_RESOURCES
)
from .core.executors.executor_state import EState as EExecutionState


__all__ = (
    "PluginAbstract",
    "EMsgType",
    "Message",
    "Credentials",
    "TesterInfoModel",
    "ModuleInfoModel",
    "PortInfoModel",
    "EProductType",
    "TesterID",
    "PIPE_EXECUTOR",
    "PIPE_RESOURCES",
    "EExecutionState",
    "PortIdentity",
    "TestParameters",
)



from __future__ import annotations

from .resource.misc import Credentials
from .resource.models.tester import TesterInfoModel
from .resource.models.module import ModuleInfoModel
from .resource.models.port import PortInfoModel

from .resource.models.types import (
    EProductType,
    TesterID,
)

__all__ = (
    "TesterID",
    "EProductType",
    "Credentials",
    "TesterInfoModel",
    "ModuleInfoModel",
    "PortInfoModel",
)



"""
Exceptions which are thrown to outside for being handled by user
"""

from .core.exceptions import (
    MultiModeError,
    TestSuiteNotExistError,
)
from .core.resources.resource.exceptions import (
    InvalidTesterTypeError,
    TesterCommunicationError,
    IsDisconnectedError,
    IsConnectedError,
    UnknownResourceError,
)
__all__ = (
    "MultiModeError",
    "TestSuiteNotExistError",
    "InvalidTesterTypeError",
    "TesterCommunicationError",
    "IsDisconnectedError",
    "IsConnectedError",
    "UnknownResourceError",
)



from __future__ import annotations

import hashlib
from typing import Type

from pydantic import (
    BaseModel,
    Field,
    SecretStr,
)
from xoa_driver import testers

from .models.types import (
    EProductType,
    TesterID,
)


class Credentials(BaseModel):
    product: EProductType
    host: str
    port: int = Field(default=22606, gt=0, lt=65535)
    password: SecretStr = SecretStr("xena")


def get_tester_inst(props: Credentials, username: str = "xoa-manager", debug=False) -> testers.GenericAnyTester | None:
    tester_type: Type[testers.GenericAnyTester] | None = {
        EProductType.VALKYRIE: testers.L23Tester,
        EProductType.CHIMERA: testers.L23Tester,
        EProductType.VANTAGE: testers.L23Tester,
        EProductType.VULKAN: testers.L47Tester,
        EProductType.SAFIRE: testers.L47Tester,
    }.get(props.product, None)
    return tester_type(
        host=props.host,
        username=username,
        password=props.password.get_secret_value(),
        port=props.port,
        enable_logging=debug,
    ) if tester_type else None


def make_resource_id(host: str, port: int) -> TesterID:
    val_bytes = f"{host}:{port}".encode("utf-8")
    return TesterID(hashlib.md5(val_bytes).hexdigest())



from __future__ import annotations
import os
from pathlib import Path
from typing import (
    Generator,
    Optional,
    Dict,
    Tuple,
    List,
)

from xoa_core.core import exceptions
from . import datasets
from . import _loader as loader


class PluginController:
    __slots__ = ("__paths", "__test_suites")

    def __init__(self) -> None:
        self.__paths: Tuple[str | Path, ...] = tuple()
        self.__test_suites: Dict[str, "datasets.PluginData"] = dict()

    def register_path(self, path: str | Path) -> None:
        """Register custom path of plugins"""
        if not os.path.isabs(path):
            path = os.path.abspath(path)
        if path in self.__paths:
            return None
        self.__paths += (path,)
        self.__init_plugins()

    def available_test_suites(self) -> List[str]:
        return list(self.__test_suites.keys())

    def suite_info(self, name: str) -> Optional[Dict]:
        suite = self.__test_suites.get(name, None)
        if not suite:
            return None
        return dict(
            test_suit=suite.meta.dict(exclude={"entry_object", "data_model"}),
            schema=suite.model_class.schema_json(),
        )

    def get_plugin_data(self, name: str) -> "datasets.PluginData":
        plugin_data = self.__test_suites.get(name, None)
        if plugin_data is None:
            raise exceptions.TestSuiteNotExistError(name)
        elif not plugin_data.meta.is_supported:
            raise exceptions.TestSuiteVersionError(name, plugin_data.meta.core_version)
        return plugin_data

    def get_plugin(self, name: str, debug: bool = False) -> "datasets.Plugin":
        plugin_data = self.get_plugin_data(name)
        return datasets.Plugin(plugin_data, debug)

    def __read_plugins(self) -> Generator["datasets.PluginData", None, None]:
        """Read plugins from provided paths."""
        for path in self.__paths:
            yield from loader.load_plugin(path)

    def __init_plugins(self) -> None:
        """Read plugins in to the local storage"""
        self.__test_suites.clear()
        self.__test_suites = dict(
            (plugin.meta.name, plugin)
            for plugin in self.__read_plugins()
        )



from typing import Callable, Optional
from enum import Enum
from functools import partialmethod


class EState(Enum):
    STOPPED = "STOPPED"
    PAUSED = "PAUSED"
    RUN = "RUN"


SenderType = Callable[[Optional[str], Optional[str]], None]


def no_sender(status: Optional[str], old_status: Optional[str]) -> None:
    return None


class ExecutorState:
    __slots__ = ("__state", "__sender", "__sender_is_set")

    def __init__(self) -> None:
        self.__state = EState.STOPPED
        self.__sender = no_sender
        self.__sender_is_set = False

    def assign_senders(self, sender: "SenderType") -> None:
        if self.__sender_is_set:
            return None
        self.__sender = sender
        self.__sender_is_set = True

    @property
    def current_state(self) -> str:
        return self.__state.value

    @property
    def is_stoped(self) -> bool:
        return self.__state is EState.STOPPED

    @property
    def is_paused(self) -> bool:
        return self.__state is EState.PAUSED

    @property
    def is_running(self) -> bool:
        return self.__state is EState.RUN

    def __set_state(self, new_state: EState) -> None:
        self.__sender(new_state.value, self.__state.value)
        self.__state = new_state

    set_stop = partialmethod(__set_state, new_state=EState.STOPPED)
    set_run = partialmethod(__set_state, new_state=EState.RUN)
    set_pause = partialmethod(__set_state, new_state=EState.PAUSED)



from __future__ import annotations

import asyncio
from dataclasses import asdict

from functools import lru_cache
from typing import (
    Any,
    Callable,
    Coroutine,
)
from xoa_driver.testers import GenericAnyTester
from xoa_core.core.utils.observer import SimpleObserver

from . import const
from . import (
    exceptions,
    misc,
)
from .models.tester import TesterModel, TesterInfoModel
from .models.types import (
    TesterID,
    StorageResource,
)


EventCallback = Callable[[TesterInfoModel, str], Coroutine[None, None, None]]


class Events:
    __slots__ = ("__observer",)

    def __init__(self, observer: SimpleObserver) -> None:
        self.__observer = observer

    def reset(self) -> None:
        self.__observer.reset()

    def on_connected(self, func: EventCallback) -> None:
        self.__observer.subscribe(const.CONNECTED, func)

    def on_disconnected(self, func: EventCallback) -> None:
        self.__observer.subscribe(const.DISCONNECTED, func)

    def on_changed(self, func: EventCallback) -> None:
        self.__observer.subscribe(const.CHANGED, func)


class Resource:
    __slots__ = ("tester", "dataset", "__observer")

    def __init__(self, credentials: misc.Credentials, *, name: str | None = None, keep_disconnected: bool | None = None) -> None:
        self.__observer: SimpleObserver[str] = SimpleObserver(pass_event=True)
        self.dataset = TesterModel(
            id=misc.make_resource_id(credentials.host, credentials.port),
            product=credentials.product,
            host=credentials.host,
            port=credentials.port,
            password=credentials.password,
            name=name or " - "
        )
        if keep_disconnected is not None:
            self.dataset.keep_disconnected = keep_disconnected
        self.tester = self.__get_tester_inst()

    def __get_tester_inst(self) -> GenericAnyTester:
        if tester_ := misc.get_tester_inst(self.credentials):
            return tester_
        raise exceptions.InvalidTesterTypeError(self.credentials)

    async def __on_tester_loose_connection(self, _) -> None:
        self.__observer.emit(const.DISCONNECTED, self.info())
        if self.keep_disconnected:
            return None
        self.tester = self.__get_tester_inst()
        for retry in range(5):
            await asyncio.sleep(delay=retry * 2)
            try:
                await self.connect()
            except Exception:
                continue
            else:
                return None
        self.dataset.keep_disconnected = True

    def __on_data_changed(self) -> None:
        self.__observer.emit(const.CHANGED, self.info())

    @property
    def is_connected(self) -> bool:
        return self.tester is not None and self.tester.session.is_online

    async def connect(self) -> None:
        if self.tester.session.is_online:
            raise exceptions.IsConnectedError(self.id)
        self.dataset.keep_disconnected = False
        try:
            await self.tester
        except Exception as e:
            raise exceptions.TesterCommunicationError(self.credentials, e) from None
        else:
            # IMPORTANT: To keep order of next functions call
            # 1 - sync which can emit CHANGED event
            # 2 - Emit CONNECTED
            # 3 - Subscribe on tester disconnected
            await self.dataset.sync(self.tester, self.__on_data_changed)
            self.__observer.emit(const.CONNECTED, self.info())
            self.tester.on_disconnected(self.__on_tester_loose_connection)

    async def disconnect(self) -> None:
        if not self.tester.session.is_online:
            raise exceptions.IsDisconnectedError(self.id)
        self.dataset.keep_disconnected = True
        await self.tester.session.logoff()
        self.tester = self.__get_tester_inst()

    async def configure(self, config: dict[str, Any]) -> None:
        raise NotImplementedError()

    def prepare_session(self, username: str, debug: bool = False) -> GenericAnyTester:
        if tester_ := misc.get_tester_inst(self.credentials, username=username, debug=debug):
            return tester_
        raise exceptions.InvalidTesterTypeError(self.credentials)

    @property
    def id(self) -> TesterID:
        return self.dataset.id

    @property
    def keep_disconnected(self) -> bool:
        return self.dataset.keep_disconnected

    @property
    def store_data(self) -> StorageResource:
        return {
            "id": self.dataset.id,
            "product": self.dataset.product,
            "host": self.dataset.host,
            "port": self.dataset.port,
            "password": self.dataset.password,
            "name": self.dataset.name,
            "keep_disconnected": self.dataset.keep_disconnected
        }

    @property
    @lru_cache
    def credentials(self) -> misc.Credentials:
        return misc.Credentials.parse_obj(self.store_data)

    def info(self) -> TesterInfoModel:
        return TesterInfoModel.parse_obj(asdict(self.dataset))

    @property
    def events(self) -> Events:
        return Events(self.__observer)



from __future__ import annotations

import asyncio
from dataclasses import (
    dataclass,
    field,
)
from typing import (
    TYPE_CHECKING,
    Callable,
    Tuple,
)

from pydantic import (
    BaseModel,
    SecretStr,
)
if TYPE_CHECKING:
    from xoa_driver import testers

from xoa_driver import utils

from .__decorator import post_notify
from .module import (
    ModuleModel,
    ModuleInfoModel,
)
from .types import (
    EProductType,
    TesterID,
)


@dataclass
class TesterModel:
    id: TesterID
    product: EProductType
    host: str
    port: int
    password: SecretStr
    name: str = " - "
    reserved_by: str = ""
    is_connected: bool = False
    modules: tuple[ModuleModel, ...] = field(default_factory=tuple)
    keep_disconnected: bool = False
    max_name_len: int = 0
    max_comment_len: int = 0
    max_password_len: int = 0
    serial_number: int = 0

    async def on_evt_reserved_by(self, _, value) -> None:
        self.reserved_by = value.username

    async def on_evt_disconnected(self, *_) -> None:
        self.is_connected = False
        self.modules = tuple()

    async def sync(self, tester: "testers.GenericAnyTester", notifier: Callable) -> None:
        tn, cpb = await utils.apply(
            tester.name.get(),
            tester.capabilities.get()
        )

        self.name = str(tn.chassis_name)
        self.reserved_by = tester.info.reserved_by
        self.is_connected = tester.session.is_online
        self.keep_disconnected = False
        self.max_name_len = cpb.max_name_len
        self.max_comment_len = cpb.max_name_len
        self.max_password_len = cpb.max_name_len
        self.serial_number = tester.info.serial_number
        self.modules = tuple(
            await asyncio.gather(*[
                ModuleModel.from_module(self.id, module, notifier)
                for module in tester.modules
            ])
        )

        tester.on_reserved_by_change(post_notify(notifier)(self.on_evt_reserved_by))
        tester.on_disconnected(self.on_evt_disconnected)


class TesterInfoModel(BaseModel):
    id: TesterID
    product: EProductType
    host: str
    port: int
    password: SecretStr
    name: str
    reserved_by: str
    is_connected: bool
    modules: Tuple[ModuleInfoModel, ...] = tuple()
    keep_disconnected: bool
    max_name_len: int
    max_comment_len: int
    max_password_len: int
    serial_number: int


from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
)
if TYPE_CHECKING:
    from xoa_driver import testers
    from xoa_core.core.generic_types import TMesagesPipe

from .pool import ResourcesPool
from .resource.facade import Resource
from .resource.misc import Credentials
from .storage import PrecisionStorage
from .types import TesterID, TesterInfoModel


class ResourcesController:
    __slots__ = ("__store", "_pool",)

    def __init__(self, msg_pipe: "TMesagesPipe", data_storage: PrecisionStorage) -> None:
        self.__store = data_storage
        self._pool = ResourcesPool(msg_pipe.transmit)

    async def start(self) -> None:
        known_testers = await self.__store.get_all()
        for credential in known_testers:
            resource = Resource(
                Credentials.parse_obj(credential),
                name=credential.get("name"),
                keep_disconnected=credential.get("keep_disconnected", False)
            )
            await self._pool.add(resource)
        failed = await self._pool.all.connect()
        for resource in failed:
            resource.dataset.keep_disconnected = True
            await self.__store.save(resource.store_data)

    async def add_tester(self, credentials: Credentials) -> TesterID:
        new_resource = Resource(credentials)  # InvalidTesterTypeError
        if await self.__store.is_registered(new_resource.id):
            return new_resource.id
        await new_resource.connect()  # TesterCommunicationError
        await self.__store.save(new_resource.store_data)
        await self._pool.add(new_resource)
        return new_resource.id

    async def remove_tester(self, id: TesterID) -> None:
        resource = await self._pool.extract(id)
        await self.__store.delete(resource.id)
        if resource.is_connected:
            await resource.disconnect()

    async def configure_tester(self, id: TesterID, config: dict[str, Any]) -> None:
        """ User Apply Changes """
        resource = self._pool.get(id)
        await resource.configure(config)

    async def list_testers_info(self) -> list[TesterInfoModel]:
        return list(self._pool.all.get_items())

    async def get_tester_info(self, tester_id: TesterID) -> TesterInfoModel:
        resource = self._pool.get(tester_id)
        return resource.info()

    async def connect(self, id: TesterID) -> None:
        resource = self._pool.get(id)
        await resource.connect()  # TesterCommunicationError, IsConnectedError
        await self.__store.save(resource.store_data)

    async def disconnect(self, id: TesterID) -> None:
        resource = self._pool.get(id)
        await resource.disconnect()  # IsDisconnectedError
        await self.__store.save(resource.store_data)

    def get_testers_by_id(self, testers_ids: Iterable[TesterID], username: str, debug: bool = False) -> dict[str, "testers.GenericAnyTester"]:
        return {
            res.id: res.prepare_session(username, debug)
            for res in self._pool.all.select(tuple(testers_ids))
        }



from .misc import Credentials


class InvalidTesterTypeError(ValueError):
    def __init__(self, props: Credentials) -> None:
        self.props = props
        self.msg = f"Can't identify Tester Type: {props.product}"
        super().__init__(self.msg)


class TesterCommunicationError(Exception):
    def __init__(self, props: Credentials, error: Exception) -> None:
        self.props = props
        self.error = error
        self.msg = f"Tester with credentials: {props} encountering communication error: {error}"
        super().__init__(self.msg)


class IsDisconnectedError(Exception):
    """Raises when tester is already disconnected."""
    def __init__(self, tester_id) -> None:
        self.tester_id = tester_id
        self.msg = f"Tester: <{self.tester_id}> is already disconnected."
        super().__init__(self.msg)


class IsConnectedError(Exception):
    """Raises when tester is already connected."""
    def __init__(self, tester_id) -> None:
        self.tester_id = tester_id
        self.msg = f"Tester: <{self.tester_id}> is already connected."
        super().__init__(self.msg)


# region Pool Exceptions

class UnknownResourceError(Exception):
    """Raises when user trying to get acces to resource which is not known."""
    def __init__(self, tester_id) -> None:
        self.tester_id = tester_id
        self.msg = f"Unknown tester of id: <{self.tester_id}>."
        super().__init__(self.msg)

# endregion



from __future__ import annotations
import asyncio
import uuid
import contextlib
from typing import (
    Set,
    AsyncGenerator
)
from xoa_core.core.utils import observer
from .pipe import MesagesPipe
from . import misc


async def _get_from_queue(queue: asyncio.Queue[misc.Message | None]) -> AsyncGenerator[misc.Message | None, None]:
    while True:
        msg = await queue.get()
        try:
            yield msg
        finally:
            queue.task_done()


class OutMessagesHandler:
    __slots__ = ("__pipes", "__senders", "__observer",)

    def __init__(self) -> None:
        self.__pipes: dict[str, MesagesPipe] = dict()
        self.__observer = observer.SimpleObserver()
        self.__observer.subscribe(misc.DISABLED, self.__on_pipe_disabled)

    def get_pipe(self, name: str) -> "MesagesPipe":
        if name in self.__pipes:
            return self.__pipes[name]
        self.__pipes[name] = pipe = MesagesPipe(
            name,
            self.__observer,
        )
        return pipe

    async def disable_pipe(self, name: str) -> None:
        if name not in self.__pipes:
            return None
        await self.__pipes[name].disable()

    def avaliable_pipes(self) -> tuple[str, ...]:
        return tuple(self.__pipes.keys())

    async def __on_pipe_disabled(self, name: str) -> None:
        del self.__pipes[name]

    @contextlib.asynccontextmanager
    async def __user_stream(self, queue: asyncio.Queue[misc.Message | None], *names: str) -> AsyncGenerator[None, None]:
        key = str(uuid.uuid4())
        pipes = (self.__pipes[name] for name in names)
        await asyncio.gather(*[pipe._add_stream(key, queue) for pipe in pipes])
        try:
            yield
        finally:
            await asyncio.gather(*[pipe._free_stream(key) for pipe in pipes])

    async def changes(self, *names: str, _filter: Set["misc.EMsgType"] | None = None) -> AsyncGenerator[misc.Message, None]:
        if not all((self.__pipes.get(name) for name in names)):
            return
        msg_queue: asyncio.Queue["misc.Message" | None] = asyncio.Queue()
        async with self.__user_stream(msg_queue, *names):
            async for msg in _get_from_queue(msg_queue):
                if msg is None:
                    break
                if _filter and msg.type not in _filter:
                    continue
                yield msg



from __future__ import annotations
import typing
from xoa_core.core.utils import observer
from xoa_core.core import exceptions
from .executor_info import ExecutorInfo
from .executor import SuiteExecutor
from ._events import Event

from xoa_core.core.generic_types import TMesagesPipe


class ExecutorsManager:

    __slots__ = ("__msg_pipe", "__executors", "__observer", "__mono")

    def __init__(self, pipe: "TMesagesPipe", mono: bool = False) -> None:
        self.__mono = mono
        self.__executors: typing.Dict[str, "SuiteExecutor"] = dict()
        self.__msg_pipe = pipe

        self.__observer = observer.SimpleObserver()
        self.__observer.subscribe(Event.STOPPED, self.__on_execution_stopped)
        self.__observer.subscribe(Event.ERROR, self.__on_execution_error)

    async def __on_execution_stopped(self, exec_id: str) -> None:
        del self.__executors[exec_id]
        self.__msg_pipe.transmit(f"Test Suite stopped: {exec_id}")

    async def __on_execution_error(self, suite_name: str, error: Exception) -> None:
        self.__msg_pipe.transmit(f"Test Suite Error: {suite_name}, {error}")

    def run(self, executor: "SuiteExecutor") -> str:
        if self.__mono and len(self.__executors) > 1:
            raise exceptions.MultiModeError()
        self.__executors[executor.id] = executor
        executor.run(self.__observer)
        self.__msg_pipe.transmit(f"Test Suite Started: {executor.id}")
        return executor.id

    def get_executors_info(self) -> list[ExecutorInfo]:
        return [
            ex.get_info()
            for ex in self.__executors.values()
        ]

    def get_state(self, exec_id: str) -> str | None:
        if ex := self.__executors.get(exec_id):
            return ex.state.current_state

    async def stop(self, exec_id: str) -> None:
        if ex := self.__executors.get(exec_id):
            await ex.stop()

    async def toggle_pause(self, exec_id: str) -> None:
        if ex := self.__executors.get(exec_id):
            await ex.toggle_pause()



import typing
from abc import (
    ABC,
    abstractmethod,
)

if typing.TYPE_CHECKING:
    from enum import Enum

    from pydantic import BaseModel
    from xoa_driver import testers as driver_testers

    from .test_suites.datasets import TestParameters


TestersType = typing.Dict[str, "driver_testers.GenericAnyTester"]
DT = typing.TypeVar("DT", bound="BaseModel")


class TransmitFunc(typing.Protocol):
    def __call__(self, msg: typing.Any, *, msg_type: "Enum", **meta) -> None: ...  # noqa: E704


class PPipeFacade(typing.Protocol):
    def __init__(self, transmit: "TransmitFunc", suite_name: str) -> None: ...  # noqa: E704

    def send_statistics(self, data: typing.Union[typing.Dict, "BaseModel"]) -> None:
        """Method used for push statistics data into the messages pipe for future distribution"""

    def send_progress(self, current: int, total: int = 100, loop: int = 0) -> None:
        """Method used for push current progress value into the messages pipe for future distribution"""

    def send_warning(self, warning: Exception) -> None:
        """Method used for push warning exceptions into the messages pipe for future distribution"""

    def send_error(self, error: Exception) -> None:
        """Method used for push error exceptions into the messages pipe for future distribution"""


class PStateConditionsFacade(typing.Protocol):
    async def wait_if_paused(self) -> None:
        """
            Block plugin execution if pause event is set.

            Notice: User must explicitly place this function at the place when pause can be happen.
        """
    async def stop_if_stopped(self) -> None:
        """
            Stop plugins execution by raising a StopPlugin exception if user call running_test_stop method.

            Notice: User must explicitly place this function at the place when pause can be happen.
        """


class PluginAbstract(ABC, typing.Generic[DT]):
    """TestSuitePlugin abstraction class, all test suite must be inherited from it."""

    __slots__ = {
        "state_conditions": """Facade contains methods which can help pause or stop Plugin execution.""",
        "xoa_out": """Facade for transmit messages to user.""",
        "testers": """Dictionary of <TESTER_ID>: <TESTER_INSTANCE>""",
        "port_identities": """PortIdentities a list where positional index is <SLOT_ID> and item is <PortIdentity>""",
        "cfg": """Test-Suite configuration model defined by plugin."""
    }

    def __init__(self, state_conditions: "PStateConditionsFacade", xoa_out: "PPipeFacade", testers: TestersType, params: "TestParameters") -> None:
        self.state_conditions = state_conditions
        """Facade contains methods which can help pause or stop Plugin execution."""
        self.xoa_out = xoa_out
        """Facade for transmit messages to user."""
        self.testers = testers
        """Dictionary of <TESTER_ID>: <TESTER_INSTANCE>"""
        self.port_identities = params.port_identities
        """PortIdentities a list where positional index is <SLOT_ID> and item is <PortIdentity>"""
        self.cfg: DT = typing.cast(DT, params.config)
        """Test configuration model defined by plugin."""
        self.prepare()

    def prepare(self) -> None:
        """Optional method used for declare plugins attributes."""
        pass

    @abstractmethod
    async def start(self) -> None:
        """Method starts test suite execution."""
        raise NotImplementedError()

    async def on_pause(self) -> None:
        """Optional method will be called and awaited when plugin receive pause event"""
        return None

    async def on_continue(self) -> None:
        """Optional method will be called and awaited when plugin receive continue event"""
        return None

    async def on_stop(self) -> None:
        """Optional method will be called and awaited when plugin receive stop event"""
        return None



from __future__ import annotations

from dataclasses import (
    dataclass,
    field,
)
from typing import (
    Any,
    Callable,
    Optional,
    Tuple,
    Type,
)
from pydantic import BaseModel
from typing_extensions import Self
from xoa_driver import (
    enums,
    ports,
    utils,
)
from .types import ModuleID, PortID
from .__decorator import post_notify


@dataclass
class PortModel:
    id: PortID
    index: int
    model: str
    reserved_by: str
    sync_status: bool | None = None
    traffic_state: enums.TrafficOnOff | None = None
    max_speed: int | None = None
    max_speed_reduction: int | None = None
    min_interframe_gap: int | None = None
    max_interframe_gap: int | None = None
    max_streams_per_port: int | None = None
    packet_limit: int | None = None
    min_packet_length: int | None = None
    max_packet_length: int | None = None
    max_header_length: int | None = None
    max_protocol_segments: int | None = None
    max_repeat: int | None = None
    can_set_autoneg: bool = False
    can_tcp_checksum: bool = False
    can_udp_checksum: bool = False
    can_micro_tpld: bool = False
    can_mdi_mdix: bool = False
    can_sync_traffic_start: bool = False
    can_fec: bool = False
    can_anlt: bool = False
    is_chimera: bool = False
    can_brr: bool = False
    speed_modes_supported: tuple["enums.PortSpeedMode", ...] = field(default_factory=tuple)
    speed_mode_current: enums.PortSpeedMode | None = None
    speed_current: int | None = None
    speed_reduction: int | None = None

    async def on_evt_traffic_state(self, _, value) -> None:
        self.traffic_state = enums.TrafficOnOff(value.on_off)

    async def on_evt_sync_status(self, _, value) -> None:
        self.sync_status = enums.SyncStatus(value.sync_status) is enums.SyncStatus.IN_SYNC

    async def on_evt_reserved_by(self, _, value) -> None:
        self.reserved_by = value.username

    @classmethod
    async def from_port(cls: Type[Self], module_id: ModuleID, port: "ports.GenericAnyPort", notifier: Callable) -> Self:
        inst = cls(
            id=PortID(f"{module_id}-{port.kind.port_id}"),
            index=port.kind.port_id,
            model=port.info.interface,
            reserved_by=port.info.reserved_by,
            sync_status=getattr(port.info, "sync_status", None),
            traffic_state=getattr(port.info, "traffic_state", None),
            **await _prepare_values(port),
        )
        port.on_reserved_by_change(post_notify(notifier)(inst.on_evt_reserved_by))
        if on_traffic_change := getattr(port, 'on_traffic_change', None):
            on_traffic_change(post_notify(notifier)(inst.on_evt_traffic_state))
        port.on_receive_sync_change(post_notify(notifier)(inst.on_evt_sync_status))
        return inst


async def _prepare_values(port: "ports.GenericAnyPort") -> dict[str, Any]:
    p_vals = dict()
    if not isinstance(port, (ports.PortL47, ports.PortL23VE)):
        if not isinstance(port, ports.PortChimera):
            ms, cur, red, tx_pl = await utils.apply(
                port.speed.mode.selection.get(),
                port.speed.current.get(),
                port.speed.reduction.get(),
                port.tx_config.packet_limit.get()
            )
            p_vals["speed_modes_supported"] = tuple(port.info.port_possible_speed_modes)
            p_vals["speed_mode_current"] = enums.PortSpeedMode(ms.mode)
            p_vals["speed_current"] = cur.port_speed
            p_vals["speed_reduction"] = None if red.ppm == -1 else red.ppm
            p_vals["packet_limit"] = tx_pl.packet_count_limit

        cpb = port.info.capabilities
        p_vals["max_speed"] = cpb.max_speed
        p_vals["max_speed_reduction"] = cpb.max_speed_reduction
        p_vals["min_interframe_gap"] = cpb.min_interframe_gap
        p_vals["max_interframe_gap"] = cpb.max_interframe_gap
        p_vals["max_streams_per_port"] = cpb.max_streams_per_port
        p_vals["min_packet_length"] = cpb.min_packet_length
        p_vals["max_packet_length"] = cpb.max_packet_length
        p_vals["max_header_length"] = cpb.max_header_length
        p_vals["max_protocol_segments"] = cpb.max_protocol_segments
        p_vals["max_repeat"] = cpb.max_repeat
        p_vals["can_set_autoneg"] = bool(cpb.can_set_autoneg)
        p_vals["can_tcp_checksum"] = bool(cpb.can_tcp_checksum)
        p_vals["can_udp_checksum"] = bool(cpb.can_udp_checksum)
        p_vals["can_micro_tpld"] = bool(cpb.can_micro_tpld)
        p_vals["can_mdi_mdix"] = bool(cpb.can_mdi_mdix)
        p_vals["can_sync_traffic_start"] = bool(cpb.can_sync_traffic_start)
        p_vals["can_fec"] = bool(cpb.can_fec)
        p_vals["can_anlt"] = bool(cpb.can_set_link_train and cpb.can_auto_neg_base_r)
        p_vals["is_chimera"] = bool(cpb.is_chimera)
        p_vals["can_brr"] = getattr(port, "is_brr_mode_supported", False)
    return p_vals


class PortInfoModel(BaseModel):
    id: PortID
    index: int
    model: str
    reserved_by: str
    sync_status: Optional[bool]
    traffic_state: Optional[enums.TrafficOnOff]
    max_speed: Optional[int]
    max_speed_reduction: Optional[int]
    min_interframe_gap: Optional[int]
    max_interframe_gap: Optional[int]
    max_streams_per_port: Optional[int]
    packet_limit: Optional[int]
    min_packet_length: Optional[int]
    max_packet_length: Optional[int]
    max_header_length: Optional[int]
    max_protocol_segments: Optional[int]
    max_repeat: Optional[int]
    can_set_autoneg: bool
    can_tcp_checksum: bool
    can_udp_checksum: bool
    can_micro_tpld: bool
    can_mdi_mdix: bool
    can_sync_traffic_start: bool
    can_fec: bool
    can_anlt: bool
    is_chimera: bool
    can_brr: bool
    speed_modes_supported: Tuple["enums.PortSpeedMode", ...]
    speed_mode_current: Optional[enums.PortSpeedMode]
    speed_current: Optional[int]
    speed_reduction: Optional[int]



PIPE_RESOURCES = "RESOURCES"
"""
Identifier of Resources Subservice for messages IO
"""
PIPE_EXECUTOR = "EXECUTOR"
"""
Identifier of Test-Suite Execution Subservice  for messages IO
"""


TESTER_ID_PATTERN = r"^[a-fA-F\d]{32}$"



from __future__ import annotations
from operator import attrgetter
from typing import (
    TYPE_CHECKING,
    Optional,
    Type,
    List,
    Dict,
    Set,
    NamedTuple,
    Any,
)
from pydantic import BaseModel
import semver

if TYPE_CHECKING:
    from ..plugin_abstract import PluginAbstract, PStateConditionsFacade, PPipeFacade

from xoa_core import __version__


class PortIdentity(BaseModel):
    tester_id: str
    module_index: int
    port_index: int

    @property
    def name(self) -> str:
        return f"P-{self.tester_id}-{self.module_index}-{self.port_index}"


class TestParameters(BaseModel):
    username: str
    port_identities: List[PortIdentity]
    config: BaseModel

    @property
    def get_testers_ids(self) -> Set[str]:
        return set(
            map(
                attrgetter("tester_id"),
                self.port_identities
            )
        )


class PluginMeta(BaseModel):
    name: str
    version: str
    core_version: str
    author: Optional[List[str]] = None
    entry_object: str
    data_model: str

    @property
    def is_supported(self) -> bool:
        return semver.match(__version__, self.core_version)


class PluginData(NamedTuple):
    """
    A test suit container.
    Contain references to starting points and metadata of the plugin
    """

    meta: PluginMeta
    entry_class: Type["PluginAbstract"]
    model_class: Type["TestParameters"]


class Plugin:
    def __init__(self, plugin_data: PluginData, debug: bool = False) -> None:
        self.plugin_data = plugin_data
        self.debug = debug

    def parse_config(self, config: Dict[str, Any]) -> None:
        self.params = self.plugin_data.model_class.parse_obj(config)  # can raise ValidationError

    def assign_testers(self, tester_getter) -> None:
        self.testers = tester_getter(
            self.params.get_testers_ids,
            self.params.username,
            self.debug
        )

    def create_test_suite(self, state_conditions: "PStateConditionsFacade", xoa_out: "PPipeFacade") -> "PluginAbstract":
        return self.plugin_data.entry_class(
            state_conditions=state_conditions,
            xoa_out=xoa_out,
            testers=self.testers,
            params=self.params,
        )


def build_test_params(_test_config: Type["BaseModel"]) -> Type["TestParameters"]:
    class TP(TestParameters):
        config: _test_config

    TP.update_forward_refs(_test_config=_test_config)
    return TP



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Type,
    Tuple,
    Optional,
)
from pydantic import BaseModel
from typing_extensions import Self
from xoa_driver import (
    enums,
    modules,
)

from .__decorator import post_notify
from .types import (
    ModuleID,
    TesterID,
)
from .port import (
    PortModel,
    PortInfoModel,
)


@dataclass
class ModuleModel:
    id: ModuleID
    index: int
    model: str
    reserved_by: str
    ports: tuple[PortModel, ...]
    serial_number: int
    name: str = " - "
    can_media_config: bool = False
    is_chimera: bool = False
    can_local_time_adjust: bool = False
    max_clock_ppm: int | None = None

    async def on_evt_reserved_by(self, _, value) -> None:
        self.reserved_by = value.username

    async def on_serial_number_change(self, _, value) -> None:
        self.serial_number = value.serial_number

    @classmethod
    async def from_module(cls: Type[Self], tester_id: TesterID, module: "modules.GenericAnyModule", notifier: Callable) -> Self:
        module_id = ModuleID(f"{tester_id}-{module.module_id}")
        inst = cls(
            id=module_id,
            index=module.module_id,
            model=module.info.model,
            reserved_by=module.info.reserved_by,
            **await _prepare_values(module),
            serial_number=(await module.serial_number.get()).serial_number,
            ports=tuple(
                await asyncio.gather(*[
                    PortModel.from_port(module_id, port, notifier)
                    for port in module.ports
                ])
            )
        )
        module.on_reserved_by_change(post_notify(notifier)(inst.on_evt_reserved_by))
        module.on_serial_number_change(post_notify(notifier)(inst.on_serial_number_change))
        return inst


async def _prepare_values(module: "modules.GenericAnyModule") -> dict[str, Any]:
    m_cpb = dict()
    if not isinstance(module, (modules.ModuleL47, modules.ModuleL47VE)):
        cpb = await module.capabilities.get()
        m_cpb["can_media_config"] = cpb.can_media_config is enums.YesNo.YES
        m_cpb["is_chimera"] = cpb.is_chimera is enums.YesNo.YES
        m_cpb["can_local_time_adjust"] = cpb.can_local_time_adjust is enums.YesNo.YES
        m_cpb["max_clock_ppm"] = cpb.max_clock_ppm

    if m_name := getattr(module, "name", None):
        mn = await m_name.get()
        m_cpb["name"] = mn.name
    return m_cpb


class ModuleInfoModel(BaseModel):
    id: ModuleID
    index: int
    model: str
    reserved_by: str
    ports: Tuple[PortInfoModel, ...]
    name: str
    can_media_config: bool
    is_chimera: bool
    can_local_time_adjust: bool
    max_clock_ppm: Optional[int]
    serial_number: int